/*
** inlinetool.c
** This utility generates inline headers for calling
** r2-based WarpOS libraries.
**
** Written by Frank Wille <frank@phoenix.owl.de> in 1998.
**
** V0.2  12-Dec-98  phx
**       inlinetool is able to concatenate prototypes, which are
**       spread over several lines.
**       Added options -p, -w, -b. Line number in error messages.
**       Bugs: Zero arguments were recognized as a single argument.
**             .extern basename is required.
** V0.1  11-Dec-98  phx
**       Continued as inline tool for r2-based libraries.
** V0.0  14-Nov-98  phx
**       File created.
*/

#include <stdio.h>
#include <ctype.h>

#define VERSION 0
#define REVISION 2
#define FDBUFSIZE 512
#define PRBUFSIZE 4096


main(int argc,char *argv[])
{
  FILE *fd,*pr;
  char fdbuf[FDBUFSIZE];
  char prbuf[PRBUFSIZE];
  char basename[64];
  char *p,*prname;
  int fdargc=0,prargc=0;
  int na,i,n,fl,prline;
  int warn = 1;
  int basename_overwrite = 0;
  int incpriv = 0;
  int fdline = 0;
  int public = 1;
  int bias = 30-6;

  if (argc < 3) {
    fprintf(stderr,"Inlinetool V%d.%d - written by Frank Wille\n\n"
            "Usage:\n%s [-w] [-p] [-b <basename>] <fd-file> <proto-file>\n",
            VERSION,REVISION,argv[0]);
    exit(10);
  }
  strcpy(basename,"unknown");

  /* parse options */
  for (i=1; i<argc; i++) {
    if (*argv[i] == '-') {
      switch ((char)tolower((unsigned)argv[i][1])) {

        case 'b':  /* overwrite base name */
          if (++i < argc) {
            strcpy(basename,argv[i]);
            basename_overwrite = 1;
          }
          else {
            fprintf(stderr,"Option -b requires library base name\n");
            exit(10);
          }
          break;

        case 'p':  /* include private functions */
          incpriv = 1;
          break;

        case 'w':  /* disable warnings */
          warn = 0;
          break;

        default:
          fprintf(stderr,"Unknown option %s ignored\n",argv[i]);
          break;
      }
    }
    else {
      if (!fdargc)
        fdargc = i;
      else if (!prargc)
        prargc = i;
      else {
        fprintf(stderr,"Too many file names\n");
        exit(10);
      }
    }
  }
  if (!fdargc || !prargc) {
    fprintf(stderr,"Missing file name\n");
    exit(10);
  }

  if ((fd = fopen(argv[fdargc],"r")) && (pr = fopen(argv[prargc],"r"))) {
    while (fgets(fdbuf,FDBUFSIZE,fd)) {
      fdline++;
      if (isalpha((unsigned)fdbuf[0])) {
        bias += 6;
        if (!public)
          continue;  /* ##private was set */

        /* sync on function name */
        p = fdbuf;
        while (*p && (*p != '('))
          ++p;
        if (!*p) {
          fprintf(stderr,"Garbage line in FD-file line %d:\n%s\n",
                  fdline,fdbuf);
          break;
        }
        *p++ = 0;  /* fdbuf is pointer to function name now */

        /* count arguments -> na */
        for (na=0; ; ) {
          while (*p && isspace((unsigned)*p) && *p!=')') ++p;
          if (*p==')')
            break;
          ++na;
          while (*p && *p!=',' && *p!=')') ++p;
          if (*p!=',')
            break;
          ++p;
        }
        if (!*p) {
          fprintf(stderr,"%s line %d: %s has a corrupted argument list\n",
                  argv[fdargc],fdline,fdbuf);
          break;
        }

        /* search function in prototype file */
        fseek(pr,0,SEEK_SET);
        prline = 0;
        fl = strlen(fdbuf);
        while (p = fgets(prbuf,PRBUFSIZE,pr)) {
          prline++;
          if ((n = strlen(prbuf) - fl) > 0) {
            for (i=0,p=prbuf; i<n; i++,p++) {
              if (!strncmp(p,fdbuf,fl))
                if (!isalnum((unsigned)*(p+fl)))  /* does name end here? */
                  break;
            }
            if (i < n)
              break;  /* function name found! */
          }
        }
        if (!p) {
          if (warn)
            fprintf(stderr,"%s not found in %s\n",fdbuf,argv[prargc]);
          continue;
        }
        prname = p;

        /* skip to end of prototype line, i.e. the ';' */
        /* concatenate prototypes spread over multiple lines, if required */
        n = 0;
        do {
          while (*p && *p!=';')
            ++p;
          if (*p == ';') {
            n = 1;
            *p = 0;  /* ok, end of prototype found, delete the ';' */
            break;
          }
        }
        while ((int)(p-prbuf) < PRBUFSIZE
               && fgets(p,PRBUFSIZE-(int)(p-prbuf),pr));

        if (!n) {
          fprintf(stderr,"Bad prototype for %s in proto file line %d:\n"
                  "%s\n",fdbuf,prline,prbuf);
          break;
        }

        /* print inline definition */
        if (prname > prbuf) {
          *(prname-1) = 0; /* beginning of function name in prototype */
          printf("%s __%s = \"",prbuf,prname);
        }
        else
          printf("__%s = \"",prbuf);  /* function has no type - uncommon! */

        printf("\\tstw\\tr2,20(r1)\\n"
               "\\t.extern\\t_%s\\n"
               "\\tlwz\\tr2,_%s(r2)\\n"
               "\\tlwz\\tr0,-%d(r2)\\n"
               "\\tmtlr\\tr0\\n"
               "\\tblrl\\n"
               "\\tlwz\\tr2,20(r1)\";\n",
               basename,basename,bias-2);

        /* print #define, which replaces the libcall */
        printf("#define %s(",fdbuf);
        for (i=1; i<=na; i++) {
          if (i != 1)
            putchar(',');
          printf("x%x",i);
        }
        printf(") __%s(",fdbuf);
        for (i=1; i<=na; i++) {
          if (i != 1)
            putchar(',');
          printf("(x%x)",i);
        }
        printf(")\n\n");
      }


      else if (!strncmp(fdbuf,"##base _",8)) {
        if (!basename_overwrite)
          sscanf(&fdbuf[8],"%s",basename);
      }

      else if (!strncmp(fdbuf,"##public",8)) {
        public = 1;
      }

      else if (!strncmp(fdbuf,"##private",9)) {
        if (!incpriv)
          public = 0;
      }

      else if (!strncmp(fdbuf,"##bias ",7)) {
        sscanf(&fdbuf[7],"%d",&bias);
        bias -= 6;
      }

      else if (!strncmp(fdbuf,"##end",5))
        break;
    }

    fclose(pr);
    fclose(fd);
  }
  else
    fprintf(stderr,"Can't open file!\n");
}
